import unittest
import hashlib
import nanotime

from dronestore.util import serial
from dronestore.model import *
from dronestore.merge import *
from dronestore.attribute import *

class PersonM(Model):
  first = StringAttribute(default="Firstname", strategy=LatestStrategy)
  last = StringAttribute(default="Lastname", strategy=LatestStrategy)
  phone = StringAttribute(default="N/A", strategy=LatestStrategy)
  age = IntegerAttribute(default=0, strategy=MaxStrategy)
  gender = StringAttribute(strategy=LatestObjectStrategy)

  def __str__(self):
    return '%s %s %s #%s age %d gender %s' % \
      (self.key, self.first, self.last, self.phone, self.age, self.gender)

class MergeTests(unittest.TestCase):

  def subtest_assert_blank_person(self, person):
    self.assertTrue(person.created.nanoseconds() == 0)
    self.assertTrue(person.committed.nanoseconds() == 0)
    self.assertTrue(person.version.isBlank)

    self.assertTrue(person.isDirty())
    self.assertFalse(person.isPersisted())
    self.assertFalse(person.isCommitted())

    self.assertEqual(person.first, 'Firstname')
    self.assertEqual(person.last, 'Lastname')
    self.assertEqual(person.phone, 'N/A')
    self.assertEqual(person.age, 0)
    self.assertEqual(person.gender, None)

  def subtest_committed(self, p, parentHash=Version.BLANK_HASH):
    self.assertFalse(p.isDirty())
    self.assertTrue(p.isCommitted())
    self.assertEqual(p.version.type, PersonM.__dstype__)
    self.assertEqual(p.version.hash, p.computedHash())
    self.assertEqual(p.version.parent, parentHash)

  def subtest_commits(self, p1, p2, diff=None):
    self.assertEqual(p1.isDirty(), p2.isDirty())
    self.assertEqual(p1.isCommitted(), p2.isCommitted())
    self.assertEqual(p1.version.type, p2.version.type)

    check = self.assertNotEqual if diff is not None else self.assertEqual
    check(p1.version, p2.version)
    check(p1.version.hash, p2.version.hash)
    check(p1.version.hash, p2.computedHash())
    check(p2.version.hash, p1.computedHash())

    if diff is None:
      diff = []

    for attr_name in p1.attributes():
      check = self.assertNotEqual if attr_name in diff else self.assertEqual

      attr1 = getattr(p1, attr_name)
      attr2 = getattr(p2, attr_name)
      check(attr1, attr2)

      attr1 = p1.version.attributeValue(attr_name)
      attr2 = p2.version.attributeValue(attr_name)
      check(attr1, attr2)

      attr1 = p1.version.attribute(attr_name)['value']
      attr2 = p2.version.attribute(attr_name)['value']
      check(attr1, attr2)


  def test_basic(self):
    a1 = PersonM('Tesla')
    a2 = PersonM('Tesla')
    self.assertEqual(a1.key, Key('/PersonM:Tesla'))
    self.assertEqual(a2.key, Key('/PersonM:Tesla'))
    self.assertEqual(a1.__dstype__, 'PersonM')
    self.assertEqual(a2.__dstype__, 'PersonM')
    self.subtest_assert_blank_person(a1)
    self.subtest_assert_blank_person(a2)

    a1.age = 52
    a1.commit()
    print 'committed a1', a1.version.hash

    a2.age = 52
    a2.commit()
    print 'committed a2', a2.version.hash

    self.subtest_committed(a1)
    self.subtest_committed(a2)
    self.subtest_commits(a1, a2)

    parentHash1 = a1.version.hash
    parentHash2 = a2.version.hash
    self.assertEqual(parentHash1, parentHash2)

    a1.first = 'Nikola'
    a1.last = 'Tesla'
    a1.phone = '7777777777'
    a1.commit()
    print 'committed a1', a1.version.hash

    a2.first = 'Nikola'
    a2.last = 'Tesla'
    a2.phone = '7777777777'
    a2.commit()
    print 'committed a2', a2.version.hash

    self.subtest_committed(a1, parentHash=parentHash1)
    self.subtest_committed(a2, parentHash=parentHash2)
    self.subtest_commits(a1, a2, diff=[])

    parentHash1 = a1.version.hash
    parentHash2 = a2.version.hash
    self.assertNotEqual(parentHash1, parentHash2)

    a1.age = 53
    self.assertTrue(a1.isDirty())
    self.assertTrue(a1.isCommitted())

    a1.commit()
    print 'committed a1', a1.version.hash

    a2.gender = 'Male'
    self.assertTrue(a2.isDirty())
    self.assertTrue(a2.isCommitted())

    a2.commit()
    print 'committed a2', a2.version.hash

    self.subtest_committed(a1, parentHash=parentHash1)
    self.subtest_committed(a2, parentHash=parentHash2)
    self.subtest_commits(a1, a2, diff=['age', 'gender'])

    parentHash1 = a1.version.hash
    a1.merge(a2)
    print 'a1 merged a2', a1.version.hash

    self.subtest_committed(a1, parentHash=parentHash1)
    self.subtest_committed(a2, parentHash=parentHash2)
    self.subtest_commits(a1, a2, diff=['age'])

    parentHash2 = a2.version.hash
    a2.merge(a1)
    print 'a2 merged a1', a2.version.hash

    self.subtest_committed(a1, parentHash=parentHash1)
    self.subtest_committed(a2, parentHash=parentHash2)
    self.subtest_commits(a1, a2)

  def test_merge_latest_object(self):
    a1 = PersonM('A')
    a2 = PersonM('A')
    a3 = PersonM('A')
    a4 = PersonM('A')

    a1.gender = 'Male'
    a2.gender = 'Female'
    a3.gender = 'Male'
    a4.gender = 'Neither'

    a1.commit()
    a2.commit()
    a3.commit()
    a4.commit()

    a3.merge(a4)
    a4.merge(a3)
    self.assertEqual(a3.gender, 'Neither')
    self.assertEqual(a4.gender, 'Neither')

    a1.merge(a2)
    a2.merge(a1)
    self.assertEqual(a1.gender, 'Female') # 2 was committed later.
    self.assertEqual(a2.gender, 'Female') # 1 was merged later, but was 2's val.

    a1.merge(a3)
    a3.merge(a1)
    self.assertEqual(a1.gender, 'Female') # 1 was committed after 3 (marge)
    self.assertEqual(a3.gender, 'Female') # 1 was committed after 3 (merge)

    a4.merge(a3)
    a3.merge(a4)
    self.assertEqual(a3.gender, 'Female')
    self.assertEqual(a4.gender, 'Female')

    self.assertEqual(a1.version.hash, a2.version.hash)
    self.assertEqual(a1.version.hash, a3.version.hash)
    self.assertEqual(a1.version.hash, a4.version.hash)


  def test_merge_latest_attribute(self):
    a1 = PersonM('A')
    a2 = PersonM('A')
    a3 = PersonM('A')
    a4 = PersonM('A')

    a1.first = 'first1'
    a2.first = 'first2'
    a3.first = 'first3'
    a4.first = 'first4'

    a1.commit()
    a2.commit()
    a3.commit()
    a4.commit()

    a4.last = 'last4'
    a3.last = 'last3'
    a2.last = 'last2'
    a1.last = 'last1'

    a1.commit()
    a2.commit()
    a3.commit()
    a4.commit()

    a1.phone = 'phone1'
    a4.phone = 'phone4'
    a3.phone = 'phone3'
    a2.phone = 'phone2'

    a1.commit()
    a2.commit()
    a3.commit()
    a4.commit()

    def check(instance, first, last, phone):
      self.assertEqual(instance.first, first)
      self.assertEqual(instance.last, last)
      self.assertEqual(instance.phone, phone)

    check(a1, 'first1', 'last1', 'phone1')
    check(a2, 'first2', 'last2', 'phone2')
    check(a3, 'first3', 'last3', 'phone3')
    check(a4, 'first4', 'last4', 'phone4')

    a3.merge(a4)
    check(a3, 'first4', 'last3', 'phone3')

    a4.merge(a3)
    check(a4, 'first4', 'last3', 'phone3')

    a1.merge(a2)
    check(a1, 'first2', 'last1', 'phone2')

    a2.merge(a1)
    check(a2, 'first2', 'last1', 'phone2')

    a1.merge(a3)
    check(a1, 'first4', 'last1', 'phone2')

    a3.merge(a1)
    check(a3, 'first4', 'last1', 'phone2')

    a4.merge(a3)
    check(a4, 'first4', 'last1', 'phone2')

    a3.merge(a4)
    check(a3, 'first4', 'last1', 'phone2')

    a1.merge(a4)
    a2.merge(a4)
    a3.merge(a4)

    check(a1, 'first4', 'last1', 'phone2')
    check(a2, 'first4', 'last1', 'phone2')
    check(a3, 'first4', 'last1', 'phone2')
    check(a4, 'first4', 'last1', 'phone2')

    self.assertEqual(a1.version.hash, a2.version.hash)
    self.assertEqual(a1.version.hash, a3.version.hash)
    self.assertEqual(a1.version.hash, a4.version.hash)


  def test_merge_max(self):
    a1 = PersonM('A')
    a2 = PersonM('A')
    a3 = PersonM('A')
    a4 = PersonM('A')

    a1.age = 11
    a2.age = 22
    a3.age = 33
    a4.age = 44

    a1.commit()
    a2.commit()
    a3.commit()
    a4.commit()

    a1.merge(a2)
    a2.merge(a1)
    self.assertEqual(a1.age, 22) # 2 was committed later.
    self.assertEqual(a2.age, 22) # 1 was merged later, but was 2's val.

    a1.merge(a3)
    a3.merge(a1)
    self.assertEqual(a1.age, 33) # 1 was committed after 3 (marge)
    self.assertEqual(a3.age, 33) # 1 was committed after 3 (merge)

    a4.merge(a3)
    a3.merge(a4)
    self.assertEqual(a3.age, 44)
    self.assertEqual(a4.age, 44)

    a1.merge(a4)
    a2.merge(a4)
    a3.merge(a4)

    a4.merge(a2)
    a4.merge(a1)
    a4.merge(a3)

    self.assertEqual(a1.age, 44)
    self.assertEqual(a2.age, 44)
    self.assertEqual(a3.age, 44)
    self.assertEqual(a4.age, 44)

    self.assertEqual(a1.version.hash, a2.version.hash)
    self.assertEqual(a1.version.hash, a3.version.hash)
    self.assertEqual(a1.version.hash, a4.version.hash)

if __name__ == '__main__':
  unittest.main()




'''Utility fields mapping Attribute classes to WTForms fields.'''


import json
import wtforms

from dronestore import Attribute, Model
from dronestore import attribute



class JSONField(wtforms.fields.TextAreaField):

  prettyprint_args = { 'sort_keys':True, 'indent':2 }
  def __init__(self, delimiter=',', prettyprint=True, default_value=None, \
      *args, **kwargs):

    self.delimiter = delimiter
    self.prettyprint = prettyprint
    self.default_value = default_value
    super(JSONField, self).__init__(*args, **kwargs)

  def joinPair(self, pair):
    return self.delimiter.join(map(str, list(pair)))

  def _value(self):
    if self.data:
      kwargs = self.prettyprint_args if self.prettyprint else {}
      return unicode(json.dumps(self.data, **kwargs))
    else:
      return unicode(json.dumps(self.default_value))

  def process_formdata(self, valuelist):
    if valuelist:
      self.data = json.loads(str(valuelist[0]))
    else:
      self.data = self.default_value




class ListField(JSONField):

  def __init__(self, *args, **kwargs):
    if 'default_value' not in kwargs:
      kwargs['default_value'] = []
    super(ListField, self).__init__(*args, **kwargs)



class DictField(JSONField):

  def __init__(self, *args, **kwargs):
    if 'default_value' not in kwargs:
      kwargs['default_value'] = {}
    super(DictField, self).__init__(*args, **kwargs)





WTFieldForAttributeMap = {
  attribute.Attribute : wtforms.fields.TextField,
  attribute.TextAttribute : wtforms.fields.TextAreaField,
  attribute.BooleanAttribute : wtforms.fields.BooleanField,
  attribute.IntegerAttribute : wtforms.fields.IntegerField,
  attribute.FloatAttribute : wtforms.fields.FloatField,
  attribute.TimeAttribute : wtforms.fields.DateTimeField,
  attribute.ListAttribute : ListField,
  attribute.DictAttribute : DictField
}




def wtfieldForAttribute(attribute):
  if isinstance(attribute, Attribute):
    attribute = attribute.__class__
  if not issubclass(attribute, Attribute):
    raise ValueError('%s is not an instance of %s' % (attribute, Attribute))

  for cls in attribute.mro():
    if cls in WTFieldForAttributeMap:
      return WTFieldForAttributeMap[cls]

  errstr = '%s passed checkes but did not lead to a valid %s'
  raise ValueError(errstr % (attribute, Attribute))



class ModelForm(wtforms.Form):

  @classmethod
  def patchWithModel(cls, model):
    for attr_name, attr in model.attributes().items():
      if not hasattr(cls, attr_name):
        fieldcls = wtfieldForAttribute(attr)
        setattr(cls, attr_name, fieldcls(label=attr_name))





WTFormForModelMap = {
  Model : ModelForm
}





def wtformForModel(model):
  if isinstance(model, str):
    model = Model.modelNamed(model)
  if isinstance(model, Model):
    model = model.__class__
  if not issubclass(model, Model):
    raise ValueError('%s is not a subclass of %s' % (model, Model))

  if model not in WTFormForModelMap:
    for cls in model.mro():
      if cls in WTFormForModelMap:
        OldForm = WTFormForModelMap[cls]
        class NewForm(OldForm):
          pass
        WTFormForModelMap[model] = NewForm
        break

  form = WTFormForModelMap[model]
  form.patchWithModel(model)

  return form




import copy
import random
import unittest

from .util import RandomGen

from dronestore.util.serial import SerialRepresentation as SR

class TestSerial(unittest.TestCase):

  def test_basic(self):
    eq = self.assertEqual

    data = {'a':1, 'b':2, 'c':3, 'd':4}

    sr1 = SR()
    for k in data:
      self.assertFalse(k in sr1)
      sr1[k] = data[k]
      self.assertTrue(k in sr1)
      self.assertEqual(sr1[k], data[k])

    sr2 = SR(copy.deepcopy(data))
    self.assertEqual(sr1, sr2)

    self.assertEqual(len(sr1), len(sr2))
    self.assertEqual(len(sr1), len(data))

    for k in data:
      self.assertTrue(k in sr1)
      self.assertTrue(k in sr2)
      del sr1[k]
      del sr2[k]
      self.assertFalse(k in sr1)
      self.assertFalse(k in sr2)

    self.assertEqual(len(sr1), 0)
    self.assertEqual(len(sr1), len(sr2))

  def __subtest_conversions(self, data):
    print 'Testing', data
    self.assertEqual(SR(data), SR(data))
    self.assertEqual(SR(data).data(), data)

  def test_conversions(self):
    self.__subtest_conversions({})
    self.__subtest_conversions({'a':1})
    self.__subtest_conversions({'a':1, 'b':2})
    self.__subtest_conversions(RandomGen.randomDict())
    self.__subtest_conversions(RandomGen.randomDict())
    self.__subtest_conversions(RandomGen.randomDict())

if __name__ == '__main__':
  unittest.main()




import os
import sys
import random
import unittest

from datastore.impl.lrucache import LRUCache
from dronestore import Key, Model, Repo, Query

from test_merge import PersonM


class TestRepo(unittest.TestCase):

  def test_simple(self):
    repo = Repo('/RepoA/', LRUCache(100))

    p = PersonM('A')
    p.first = 'A'
    p.last = 'B'
    p.commit()

    self.assertFalse(repo.contains(p.key))
    self.assertEqual(repo.get(p.key), None)

    repo.put(p)
    self.assertTrue(repo.contains(p.key))
    self.assertEqual(repo.get(p.key), p)
    for i in range(0, 100):
      repo.delete(p.key)
      self.assertEqual(repo.get(p.key), None)
      self.assertFalse(repo.contains(p.key))

      repo.put(p)
      self.assertEqual(repo.get(p.key), p)
      self.assertTrue(repo.contains(p.key))


    p2 = PersonM(p.version)
    self.assertEqual(p2, repo.get(p2.key))

    p2.first = 'B'
    p2.commit()

    self.assertNotEqual(p, p2)
    self.assertNotEqual(p.first, p2.first)
    self.assertNotEqual(p2, repo.get(p2.key))

    p2 = repo.merge(p2)

    self.assertEqual(p2, repo.get(p2.key))

    q = Query(PersonM).filter('first', '=', 'B')
    res = list(repo.query(q))

    self.assertTrue(len(res) == 1)
    self.assertEqual(p2, res[0])


  def test_stress(self):
    num_repos = 5
    num_people = 10

    repos = []
    for i in range(0, num_repos):
      repos.append(Repo('/Repo%s' % i, LRUCache(num_people)))
      print 'Created repo ', repos[-1].repoid


    people = []
    for i in range(0, num_people):
      p = PersonM('person%s' % i)
      p.first = 'first%d:' % i
      p.last = 'last%d:' % i
      p.phone = '%d:' % i
      p.gender = '%d:' % i
      p.age = 0
      p.commit()

      d = random.choice(repos)
      d.put(p)
      print 'Created person: ', p.key

    def randomPerson(repo):
      i = random.randint(0, num_people - 1)
      return repo.get(Key('/PersonM:person%s' % i))

    def updateField(field):
      d = random.choice(repos)
      p = randomPerson(d)
      if p is None:
        return

      if field == 'age':
        p.age += 1
      else:
        setattr(p, field, field + str(i))
      p.commit()
      d.merge(p)

    def shuffle():
      d1, d2 = random.sample(repos, 2)
      p = randomPerson(d1)
      if p is None:
        return

      d2.merge(p)


    for i in range(num_people * 10):
      updateField('first')
      updateField('last')
      updateField('phone')
      updateField('age')
      updateField('gender')

      shuffle()
      shuffle()
      shuffle()
      shuffle()
      shuffle()

    for d in repos:
      print 'Repo: ', d.repoid
      for i in range(num_people):
        key = Key('/PersonM:person%s' % i)

        p = d.get(key)
        if p is None:
          print key, 'not found'
        else:
          print p

    for i in range(num_people):
      key = Key('/PersonM:person%s' % i)
      p = repos[0].get(key)
      for d in repos:
        p = d.merge(p)

      # doule pass.
      for d in repos:
        p = d.merge(p)
        o = d.get(p.key)
        self.assertEqual(p, o)
        self.assertEqual(p.first, o.first)
        self.assertEqual(p.last, o.last)
        self.assertEqual(p.phone, o.phone)
        self.assertEqual(p.age, o.age)
        self.assertEqual(p.gender, o.gender)
        self.assertEqual(p.version, o.version)

        print 'In order:'
        last = None
        for person in d.query(Query(PersonM).order('key')):
          if last:
            self.assertTrue(str(person.key) > last)
          last = str(person.key)
          print person

        print 'Reverse:'
        last = None
        for person in d.query(Query(PersonM).order('-key')):
          if last:
            self.assertTrue(str(person.key) < last)
          last = str(person.key)
          print person

        result = d.query(Query(PersonM, limit=3).order('-key'))
        self.assertTrue(len(list(result)) == 3)

if __name__ == '__main__':
  unittest.main()



import time
import datetime
import unittest
import nanotime

class TestTime(unittest.TestCase):

  def test_construction(self):

    def eq(time1, time2):
      self.assertEqual(time1._ns, time2._ns)
      self.assertEqual(time1.nanoseconds(), time2.nanoseconds())
      self.assertEqual(time1.microseconds(), time2.microseconds())
      self.assertEqual(time1.milliseconds(), time2.milliseconds())
      self.assertEqual(time1.seconds(), time2.seconds())
      self.assertEqual(time1.minutes(), time2.minutes())
      self.assertEqual(time1.hours(), time2.hours())
      self.assertEqual(time1.days(), time2.days())
      self.assertEqual(time1.timestamp(), time2.timestamp())
      self.assertEqual(time1.datetime(), time2.datetime())
      self.assertEqual(time1.unixtime(), time2.unixtime())

    def close(x, y, epsilon=1e-6):
        return abs(x - y) < epsilon

    for i in range(0, 50):
      # basic
      eq(nanotime.seconds(1), nanotime.nanoseconds(1e9))
      eq(nanotime.seconds(1), nanotime.microseconds(1e6))
      eq(nanotime.seconds(1), nanotime.milliseconds(1e3))
      eq(nanotime.seconds(1), nanotime.seconds(1))
      eq(nanotime.seconds(1), nanotime.minutes(1.0/60))
      eq(nanotime.seconds(1), nanotime.hours(1.0/3600))
      eq(nanotime.seconds(1), nanotime.days(1.0/(3600 * 24)))

      # timestamp
      ts1 = time.time()
      ts2 = nanotime.timestamp(ts1).timestamp()
      eq(nanotime.timestamp(ts1), nanotime.timestamp(ts1))
      self.assertTrue(abs(ts2 - ts1) < 10e-6)

      # datetime
      dt1 = datetime.datetime.now()
      dt2 = nanotime.datetime(dt1).datetime()
      eq(nanotime.datetime(dt1), nanotime.datetime(dt1))
      self.assertTrue(abs(dt1 - dt2) < datetime.timedelta(microseconds=1))

  def __subtest_arithmetic(self, start, extra):
    eq = self.assertEqual

    start = float(start)
    extra = float(extra)
    t1 = nanotime.seconds(start)
    eq(nanotime.seconds(start + extra), t1 + nanotime.nanotime(extra * 1e9))
    eq(nanotime.seconds(start + extra), t1 + nanotime.seconds(extra))
    eq(nanotime.seconds(start - extra), t1 - nanotime.nanotime(extra * 1e9))
    eq(nanotime.seconds(start - extra), t1 - nanotime.seconds(extra))
    eq(nanotime.seconds(start * extra), t1 * nanotime.nanotime(extra))
    eq(nanotime.seconds(start * extra), t1 * nanotime.nanoseconds(extra))
    eq(nanotime.seconds(start / extra), t1 / nanotime.nanotime(extra))
    eq(nanotime.seconds(start / extra), t1 / nanotime.nanoseconds(extra))

    self.assertTrue(nanotime.seconds(start + extra) > t1)
    self.assertTrue(nanotime.seconds(start - extra) < t1)

    t2 = nanotime.seconds(start + extra)
    self.assertTrue(t2  > nanotime.nanotime(0))
    self.assertTrue(nanotime.nanotime(0) < t2)

  def test_arithmetic(self):
    for start in range(0, 10000, 1000):
      for extra in range(1, 20, 5):
        self.__subtest_arithmetic(start, extra)


if __name__ == '__main__':
  unittest.main()




import datetime
import hashlib
import uuid
import nanotime
import copy

from datastore import Key

from .util import serial
from .util import fasthash

import merge


class UnregisteredModelError(ValueError):
  pass

class DuplicteModelError(ValueError):
  pass

class DuplicateAttributeError(ValueError):
  pass

class InternalValueError(ValueError):
  pass



class classproperty(object):
  '''Implements both @property and @classmethod behavior.'''
  def __init__(self, getter):
    self.getter = getter
  def __get__(self, instance, owner):
    return self.getter(instance) if instance else self.getter(owner)



class Version(object):
  ''' A version is one snapshot of a particular object's values.

  Versions have an associated hash (sha1). Their hash determines uniqueness of
  the object snapshot. Versions are used as snapshot 'containers,' including
  all of the data of the particular object snapshot.

  The current implementation does not use incremental changes, as the entire
  version history of each object is not tracked.
  '''
  BLANK_HASH = '0000000000000000000000000000000000000000'
  REP_FIELDS = ['key', 'hash', 'parent', 'created', 'committed', 'attributes', \
    'type']

  def __init__(self, keyOrRepresentation):
    serialRep = None
    key = None

    if isinstance(keyOrRepresentation, serial.SerialRepresentation):
      serialRep = keyOrRepresentation
    elif isinstance(keyOrRepresentation, Key):
      key = keyOrRepresentation
    else:
      raise ValueError('No serial representation or key provided.')


    if serialRep is None:
      serialRep = serial.SerialRepresentation()
      serialRep['key'] = str(key)
      serialRep['hash'] = self.BLANK_HASH
      serialRep['parent'] = self.BLANK_HASH
      serialRep['created'] = 0
      serialRep['committed'] = 0
      serialRep['attributes'] = {}
      serialRep['type'] = ''

    for req in self.REP_FIELDS:
      if req not in serialRep:
        raise ValueError('serial representation does not include %s' % req)

    if serialRep['created'] > serialRep['committed']:
      raise ValueError('serial representation implies created after committed')
    if serialRep['created'] < 0:
      raise ValueError('serial representation implies created before 0')

    self._serialRep = serialRep

  @property
  def key(self):
    return Key(self._serialRep['key'])

  @property
  def hash(self):
    return self._serialRep['hash']

  @property
  def type(self):
    return self._serialRep['type']

  @property
  def isBlank(self):
    return self.hash == self.BLANK_HASH

  def shortHash(self, length=6):
    return self.hash[0:length]

  @property
  def committed(self):
    return nanotime.nanotime(self._serialRep['committed'])

  @property
  def created(self):
    return nanotime.nanotime(self._serialRep['created'])

  @property
  def parent(self):
    return self._serialRep['parent']

  @property
  def serialRepresentation(self):
    return self._serialRep

  def attribute(self, name):
    try:
      return self._serialRep['attributes'][name]
    except KeyError:
      raise KeyError('No attribute %s in version' % name)

  def attributeValue(self, name):
    return self.attributeMetaData(name, 'value')

  def attributeMetaData(self, name, meta):
    attr = self.attribute(name) # outside the try to propagate up attr errors
    try: # try catch here for the usual case.
      return attr[meta]
    except KeyError:
      errstr = 'No attribute metadata \'%s\' in this version. %s'
      raise KeyError(errstr % (meta, self._serialRep.data()))

  def __getitem__(self, name):
    return self.attribute(name)

  def __eq__(self, other):
    if isinstance(other, Version):
      return self.hash == other.hash and self.key == other.key
    raise TypeError('other is not of type %s' % Version)

  def __ne__(self, other):
    return not self.__eq__(other)

  def __hash__(self):
    return fasthash.hash(self.hash)

  def __str__(self):
    return '<%s %s version %s>' % (self.type, self.key, self.hash)


attribute = None

def _initialize_attributes(cls, name, bases, attrs):
  '''This function initializes attributes (and handles name collisions).
  Attribute binding follows the model that property binding does in the Google
  App Engine.
  '''
  global attribute
  if not attribute:
    import attribute


  cls._attributes = {}
  defined_attrs = {}

  # this walks the bases to find which class added the given attr.
  def get_attr_source(cls, attr):
    for src_cls  in cls.mro():
      if attr in src_cls.__dict__:
        return src_cls

  # Gather all the ds attributes from all the bases.
  for base in bases:
    if hasattr(base, '_attributes'):
      keys = set(base._attributes.keys())
      dupe_keys = set(defined_attrs.keys()) & keys
      for dupe_key in dupe_keys:
        old_source = defined_attrs[dupe_key]
        new_source = get_attr_source(base, dupe_prop_name)
        if old_source != new_source:
          raise DuplicateAttributeError(
              'Duplicate attribute, %s, is inherited from both %s and %s.' %
              (dupe_prop_name, old_source.__name__, new_source.__name__))

      keys -= dupe_keys
      if keys:
        defined_attrs.update(dict.fromkeys(keys, base))
        cls._attributes.update(base._attributes)

  # add the ds attributes from this class.
  for attr_name in attrs.keys():
    attr = attrs[attr_name]
    if isinstance(attr, attribute.Attribute):
      # reserved word check here.
      if attr_name in defined_attrs:
        raise DuplicateAttributeError('Duplicate attribute: %s' % attr_name)
      defined_attrs[attr_name] = cls
      cls._attributes[attr_name] = attr
      attr._attr_config(cls, attr_name)


REGISTERED_MODELS = {}

class ModelMeta(type):
  '''This is the meta class for Model.
  It sets up model attributes and registers the object.
  '''
  def __init__(cls, name, bases, attrs):
    super(ModelMeta, cls).__init__(name, bases, attrs)

    _initialize_attributes(cls, name, bases, attrs)

    type_name = cls.__dstype__
    if type_name == 'Model':
      cls.__dstype__ = cls.__name__
      type_name = cls.__dstype__

    if type_name in REGISTERED_MODELS and REGISTERED_MODELS[type_name] != cls:
      raise DuplicteModelError('Duplicate model registered: %s' % type_name)
    REGISTERED_MODELS[type_name] = cls




class Model(object):
  '''Model'''
  __metaclass__ = ModelMeta
  __dstype__ = 'Model'

  def __init__(self, keyNameOrVersion, parentKey=None):
    '''Initializes the model by reconstructing from version or blank state.'''

    if isinstance(keyNameOrVersion, Version):
      self._initialize_version(keyNameOrVersion)

    elif isinstance(keyNameOrVersion, Key) or isinstance(keyNameOrVersion, str):
      self._initialize_new(keyNameOrVersion, parentKey)

    else:
      raise ValueError('No key or version provided.')


  def _initialize_new(self, key_name, parentKey):
    '''Initializes with blank state (only key)'''

    key_name = str(key_name)
    if '/' in key_name:
      raise ValueError('Key name %s includes slashes. It must not.' % key_name)

    key = Key('/%s:%s' % (self.__dstype__, key_name))
    if parentKey:
      key = parentKey.child(key)

    for attr in self.attributes().values():
      attr.__set__(self, attr.default_value(), default=True)


    self._key = key
    self._version = Version(key)
    self._updated = None
    self._isDirty = True
    self._isPersisted = False

  def _initialize_version(self, version):
    '''Initializes from stored version data'''
    #FIXME(jbenet) consider moving this to Version class...

    if version.type != self.__class__.__dstype__:
      raise ValueError('Type name provided does not match.')

    for attr in self.attributes().values():
      try:
        value = copy.copy(version.attributeValue(attr.name))
      except KeyError:
        value = attr.default_value()
        if not value and attr.required:
          raise
      attr.__set__(self, value)


    self._key = version.key
    self._version = version
    self._isDirty = False
    self._isPersisted = True

  @classproperty
  def key(cls_or_self):
    '''The key associated with this model/instance.
    This is a classproperty in order to have:

        >>> Model.key
        Key('/Model')
        >>> Model('instance').key
        Key('/Model:instance)

    '''
    if isinstance(cls_or_self, type):
      return Key(cls_or_self.__dstype__)
    return cls_or_self._key


  @property
  def created(self):
    '''When this object was created.'''
    return self._version.created

  @property
  def committed(self):
    '''When this object was last committed.'''
    return self._version.committed

  @property
  def version(self):
    '''The current version of this object.'''
    return self._version

  def isPersisted(self):
    return self._isPersisted

  def isCommitted(self):
    return not self._version.isBlank

  def isDirty(self):
    return self._isDirty

  @classmethod
  def attributes(cls):
    '''Returns a dictionary of all the attributes defined for this model.'''
    return dict(cls._attributes)

  def attributeValues(self):
    '''Returns the attribute values of this model.'''
    return dict([(a, getattr(self, a)) for a in self.attributes()])

  def validate(self):
    '''Validates the instance attributes, ensuring invariants hold.

    Override this method (remember to call base classes' validate) to intrude
    commit-time checks that should prevent storing incorrect values.
    '''

    if self.committed < self.created:
      raise ValueError('Internal commit time is earlier than creation time')

    for attr_name, attr in self.attributes().iteritems():
      attr.validate(getattr(self, attr_name))


  def computedHash(self):
    buf = '%s,%s,' % (self._key, self.__dstype__)
    for attr_name, attr in self.attributes().iteritems():
      buf += '%s=%s,' % (attr_name, attr.rawData(self))
    return hashlib.sha1(buf).hexdigest()

  def commit(self):
    '''Committing a version creates a snapshot of the current changes.'''

    # this is actually broken for collection attributes:
    # if not self.isDirty():
    #   return # nothing to commit

    self.validate()

    sr = serial.SerialRepresentation()
    sr['hash'] = self.computedHash()
    if sr['hash'] == self._version.hash:
      self._isDirty = False
      return # false alarm, nothing to commit.

    sr['key'] = str(self.key)
    sr['type'] = self.__dstype__
    sr['parent'] = self._version.hash
    sr['created'] = self._version.created.nanoseconds()
    sr['committed'] = nanotime.now().nanoseconds()
    sr['attributes'] = {}

    if sr['created'] == 0: # from blank version
      sr['created'] = sr['committed']

    for attr_name, attr in self.attributes().iteritems():
      sr['attributes'][attr_name] = serial.clean(attr.rawData(self))

    self._version = Version(sr)

    self._isPersisted = True
    self._isDirty = False

  def merge(self, other):
    if isinstance(other, Version):
      merge.merge(self, other)
    elif isinstance(other, Model):
      merge.merge(self, other.version)
    else:
      raise TypeError('Expected instance of %s or %s' % \
        (Version, self.__class__))

  def __eq__(self, o):
    '''Check equality with another model of this kind.'''
    if not isinstance(o, self.__class__):
      return False

    if self.version != o.version:
      return False

    # if there are no changes, checking versions is enough
    if not self._isDirty and not o._isDirty:
      return True

    # we must check every attribute
    for attr in self.attributes():
      if getattr(self, attr) != getattr(o, attr):
        return False

    return True

  def __ne__(self, other):
    return not self.__eq__(other)

  @classmethod
  def modelNamed(cls, name):
    try:
      return REGISTERED_MODELS[name]
    except KeyError:
      errstr = 'Model %s is not registered. Registered models are: %s'
      raise UnregisteredModelError(errstr % (name, REGISTERED_MODELS))

  @classmethod
  def from_version(cls, version):
    return cls.modelNamed(version.type)(version)




from datastore.query import Query as DatastoreQuery
from datastore.query import Filter, Order

from model import Key, Version, Model
from util import serial


def _object_getattr(obj, field):
  '''Aggressively tries to determine a value in `obj` for identifier `field`'''

  value = None

  # check whether this key is an attribute (Model, etc)
  if hasattr(obj, field):
    value = getattr(obj, field)

  # if not, perhaps it is an item (raw dicts, etc)
  elif field in obj:
    value = obj[field]

  # if not, perhaps it is an attributeValue (Version)
  elif hasattr(obj, 'attributeValue'):
    value = obj.attributeValue(field)

  # if not, perhaps it is an attribute (SerialRepresentations)
  elif 'attributes' in obj and field in obj['attributes']:
    value = obj['attributes'][field]['value']

  # return whatever we've got.
  return value




class Query(DatastoreQuery):

  def __init__(self, key, *args, **kwargs):

    if hasattr(key, '__dstype__'):
      key = Key(key.__dstype__)

    if isinstance(key, str):
      key = Key(key)

    super(Query, self).__init__(key, *args, **kwargs)

  def model(self):
    '''Returns the Model class associated to this query.'''
    return Model.modelNamed(self.key.name)

  object_getattr = staticmethod(_object_getattr)



def allinstances(cls, droneOrDatastore):
  '''Returns the result of querying `droneOrDatastore` with type `cls`'''
  if not issubclass(cls, Model):
    raise TypeError('cls must derive from %s' % Model)
  return droneOrDatastore.query(Query(cls))

Model.all = classmethod(allinstances)



class InstanceIterator(object):
  '''Wraps an iterator to convert Version SerialRepresentations to instances.
  Used mainly around queries to ensure iterating over the result iterator will
  return instances, not raw version data.
  '''

  def __init__(self, iterable):
    self.iter = iter(iterable)

  def __iter__(self):
    return self

  def next(self):
    '''Returns an instance of the version represented by the next object.'''
    if self.iter is None:
      raise StopIteration

    # if it returns none, return None as well. None is not necessarily the end.
    next = self.iter.next()
    if next is None:
      return None

    # if it is a dictionary, assume raw serial representation
    if isinstance(next, dict):
      next = serial.SerialRepresentation(next)

    # if it is a serialRepresentation, turn it into a Version
    if isinstance(next, serial.SerialRepresentation):
      next = Version(next)

    # if it is a Version, turn it into a Model
    if isinstance(next, Version):
      next = Model.from_version(next)

    # return whatever it is we have!
    return next






import nanotime

def merge(instance, version):
  if instance.isDirty():
    raise ValueError('Cannot merge dirty instance.')

  if not instance.isCommitted():
    raise ValueError('Cannot merge uncommitted instance.')

  mergeData = {}
  for attr in instance.attributes().values():
    rawData = attr.mergeStrategy.merge(instance.version, version)
    if rawData: # none value means no change, i.e. keep the local attribute.
      mergeData[attr.name] = rawData

  if not mergeData:
    return # nothing changed.

  # merging checks out, actually make the changes.
  for attrname, rawData in mergeData.iteritems():
    attr = instance.attributes()[attrname]
    attr.setRawData(instance, rawData)

  instance.commit()


class MergeDirection:
  '''MergeDirection represents an enumeration to identify which side to keep.'''
  Local, Remote, Merge = range(1,4)

class MergeStrategy(object):
  '''A MergeStrategy represents a unique way to decide how the two values of a
  particular attributes merge together.

  MergeStrategies are meant to enforce a particular rule that helps ensure
  application semantics regarding attributes changed in multiple nodes.

  MergeStrategies can store state in the object (e.g. a timestamp). If so,
  MergeStrategies must set the REQUIRES_STATE class variable to True.
  '''

  REQUIRES_STATE = False


  def __init__(self, attribute):
    self.attribute = attribute

  def merge(self, local_version, remote_version):
    '''merges this attribute in two different versions.'''
    raise NotImplementedError('No implementation for %s.merge()', \
      self.__class__.__name__)

  def setAttribute(self, instance, rawData, default=False):
    '''Called whenever this particular attribute is set to a new value.'''
    pass

  def _attribute_data(self, version):
    '''Returns the attribute data for this version or None'''
    try:
      attr = version.attribute(self.attribute.name)
    except KeyError:
      attr = None
    return attr


class LatestObjectStrategy(MergeStrategy):
  '''LatestObjectStrategy merges attributes based solely on objects' timestamp.
  In essence, the most recently written object wins.

  This Strategy stores no additional state.
  '''

  def merge(self, local_version, remote_version):
    if remote_version.committed > local_version.committed:
      return remote_version.attribute(self.attribute.name)
    return None





class LatestStrategy(MergeStrategy):
  '''LatestStrategy merges attributes based solely on timestamp. In essence, the
  most recently written attribute wins.

  This Strategy stores its state like so:
  { 'updated' : nanotime.nanotime, 'value': attrValue }

  A value with a timestamp will be preferred over values without.
  '''

  REQUIRES_STATE = True

  def merge(self, local_version, remote_version):

    attr_local = self._attribute_data(local_version)
    attr_remote = self._attribute_data(remote_version)

    # if no timestamp found in remote. we're done!
    if not attr_remote or 'updated' not in attr_remote:
      return None

    # since other side has a timestamp, if we don't, take theirs.
    if not attr_local or 'updated' not in attr_local:
      return attr_remote

    # if we havent decided (both have timestamps), compare timestamps
    if attr_remote['updated'] > attr_local['updated']:
      return attr_remote
    return None # no change. keep local

  def setAttribute(self, instance, rawData, default=False):
    '''Called whenever this particular attribute is set to a new value.'''
    # update the update metadata to reflect the current time.
    if default:
      rawData['updated'] = 0
    else:
      rawData['updated'] = nanotime.now().nanoseconds()





class MaxStrategy(MergeStrategy):
  '''MaxStrategy merges attributes based solely on comparison. In essence, the
  larger value is picked.

  This Strategy stores no additional state.
  '''

  def merge(self, local_version, remote_version):

    attr_local = self._attribute_data(local_version)
    attr_remote = self._attribute_data(remote_version)

    if not attr_remote:
      return None

    if not attr_local:
      return attr_remote

    if attr_remote['value'] > attr_local['value']:
      return attr_remote
    return None # no change. keep local







from model import Key, Version, Model
from query import Query, InstanceIterator
from datastore import Datastore, DictDatastore
from .util.serial import SerialRepresentation

class Repo(object):
  '''Repo represents the logical unit of storage in dronestore.
  Each repo consists of a datastore (or set of datastores) and an id.
  '''

  #FIXME(jbenet): remove DictDatastore as a default?
  def __init__(self, repoid, store=DictDatastore()):
    '''Initializes drone with given id and datastore.'''
    if not isinstance(repoid, Key):
      repoid = Key(repoid)
    if not isinstance(store, Datastore):
      raise ValueError('store must be an instance of %s' % Datastore)

    self._repoid = repoid
    self._store = store

  # deprecated
  @property
  def droneid(self):
    '''This repo's identifier.'''
    return self._repoid

  @property
  def repoid(self):
    '''This repo's identifier.'''
    return self._repoid

  def __str__(self):
    return '<dronestore.drone.Repo object at %s %s>' % (id(self), self.repoid)

  @classmethod
  def _cleanVersion(cls, parameter):
    '''Extracts the version from input.'''
    if isinstance(parameter, Version):
      return parameter
    elif isinstance(parameter, Model):
      if parameter.isDirty():
        raise ValueError('cannot store entities with uncommitted changes')
      return parameter.version

    raise TypeError('expected input of type %s or %s' % (Version, Model))


  def put(self, versionOrEntity):
    '''Stores the current version of `entity` in the datastore.'''
    version = self._cleanVersion(versionOrEntity)
    self._store.put(version.key, version.serialRepresentation.data())
    return versionOrEntity


  def get(self, key):
    '''Retrieves the current entity addressed by `key`'''
    if not isinstance(key, Key):
      raise ValueError('key must be of type %s' % Key)

    # lookup the key in the datastore
    data = self._store.get(key)
    if data is None:
      return data

    # handle the data. if any conversion fails, propagate the exception up.
    serialRep = SerialRepresentation(data)
    version = Version(serialRep)
    return Model.from_version(version)


  def merge(self, newVersionOrEntity):
    '''Merges a new version of an instance with the current one in the store.'''

    # get the new version
    new_version = self._cleanVersion(newVersionOrEntity)

    # get the instance
    key = new_version.key
    curr_instance = self.get(key) #THINKME(jbenet): try contains first?

    # brand new version. just store it.
    if curr_instance is None:
      self.put(new_version)
      return Model.from_version(new_version)

    # NOTE: semantically, we must merge into the current instance in the repo
    # so that merge strategies favor the incumbent version.
    curr_instance.merge(new_version)

    # store it back
    self.put(curr_instance)
    return curr_instance


  def contains(self, key):
    '''Returns whether the datastore contains the entity addressed by `key`.'''
    if not isinstance(key, Key):
      raise ValueError('key must be of type %s' % Key)

    return self._store.contains(key)


  def delete(self, key):
    '''Deletes the entity addressed by `key` from the datastore.'''
    if not isinstance(key, Key):
      raise ValueError('key must be of type %s' % Key)

    self._store.delete(key)

  def query(self, query):
    '''Queries the datastore for objects matching `query`.'''
    return InstanceIterator(self._store.query(query))




from types import DictType
from types import StringType



class ProtocolError(Exception):
  __slots__ = ('kind', 'message')
  def __init__(self, call, message):
    self.kind = 'Protocol (%d)' % call
    self.message = message

class RPC(object):
  __slots__ = ('kind', 'args')

  def __repr__(self):
    return 'RPC %s (%s)' % (self.kind, str(self.args))

  def __str__(self):
    return 'RPC %s (%s)' % (self.kind, str(self.args))

  def handle(self):
    raise NotImplementedError('No implementation for RPC ')



'''
Bottle-dronestore is a plugin that integrates dronestore with your Bottle
application. It passes a given repo to the route callback. It is heavily based
on the bottle-sqlite.

To automatically detect routes that need a repo, the plugin
searches for route callbacks that require a `repo` keyword argument
(configurable) and skips routes that do not. This removes any overhead for
routes that don't need a repo connection.

Usage Example::

    import bottle
    from dronestore.ext import dsbottle

    app = bottle.Bottle()
    plugin = sqlite.Plugin(dbfile='/tmp/test.db')
    app.install(plugin)

    @app.route('/show/:item')
    def show(item, db):
        row = db.execute('SELECT * from items where name=?', item).fetchone()
        if row:
            return template('showitem', page=row)
        return HTTPError(404, "Page not found")
'''

__author__ = 'Juan Batiz-Benet'
__email__ = 'juan@benet.ai'
__version__ = '0.1'
__license__ = 'MIT'

import inspect
try:
  from bottle import PluginError
except:
  class PluginError(Exception):
    pass


class DronestoreBottlePlugin(object):
  ''' This plugin passes a dronestore repo handle to route callbacks
      that accept a `repo` keyword argument. If a callback does not expect
      such a parameter, no repo is passed. You can override the repo
      settings on a per-route basis. '''

  name = 'dronestore'
  api = 2

  def __init__(self, repo=None, drone=None, keyword='repo'):
    repo = repo or drone # deprecate drone
    self.repo = repo
    self.keyword = keyword

  # deprecate drone
  @property
  def drone(self):
    return self.repo

  def __repr__(self):
    return '<%s %s=%s>' % \
      (DronestoreBottlePlugin, self.keyword, self.repo)

  def setup(self, app):
    ''' Make sure that other installed plugins don't affect the same
        keyword argument.'''
    for other in app.plugins:
      if not isinstance(other, DronestoreBottlePlugin):
        continue

      if other.keyword == self.keyword:
        errstr = 'dronestore bottle plugin keyword conflict between %s and %s'
        raise PluginError(errstr % (repr(self), repr(other)))

  def apply(self, callback, route):
    # Override global configuration with route-specific values.
    repo = self.repo
    if hasattr(route, 'config'):
      repo = route.config.get(self.keyword) or repo

    if not repo:
      raise PluginError('No repo specified for %s' % repr(self))

    # Test if the original callback accepts a keyword argument for this repo
    # Ignore it if it does not need a repo handle.
    if hasattr(route, 'callback'):
      args = inspect.getargspec(route.callback)[0]

      if self.keyword not in args:
        return callback

    def wrapper(*args, **kwargs):
      kwargs[self.keyword] = repo
      return callback(*args, **kwargs)

    # Replace the route callback with the wrapped one.
    return wrapper


Plugin = DronestoreBottlePlugin




import unittest
import nanotime

from dronestore.model import *
from dronestore.attribute import *


class AttributeTests(unittest.TestCase):

  def subtest_attribute(self, attrtype, **kwds):
    error = lambda: Attribute(merge_strategy='bogus')
    self.assertRaises(TypeError, error)

    a = attrtype(name='a', **kwds)
    self.assertEqual(a.name, 'a')
    self.assertEqual(a._attr_name(), '_a')
    self.assertEqual(a.default, None)
    self.assertEqual(a.default_value(), None)
    self.assertTrue(a.empty(None))
    self.assertFalse(a.required)
    self.assertTrue(isinstance(a.mergeStrategy, merge.MergeStrategy))

    # setup the binding to an object
    class Attributable(object):
      pass

    m = Attributable()
    m._attributes = {'a':a}
    a._attr_config(Attributable, 'a')

    self.assertEqual(a._attr_name(), '_a')
    self.assertEqual(m._attributes['a'], a)

    def testSet(value, testval=None):
      a.__set__(m, value)
      if not testval:
        testval = value
      self.assertEqual(a.__get__(m, object), testval)

    return testSet

  def test_basicattributes(self):
    test = self.subtest_attribute(Attribute)
    test(5, '5')
    test(5.2, '5.2')
    test(self, str(self))
    test('5')
    test(None)

    test = self.subtest_attribute(StringAttribute)
    test(5, '5')
    test(5.2, '5.2')
    test(self, str(self))
    test('5')
    test(None)
    self.assertRaises(ValueError, test, '5\n\n\nfdsijhfdiosahfdsajfdias')

    test = self.subtest_attribute(StringAttribute, multiline=True)
    test(5, '5')
    test(5.2, '5.2')
    test(self, str(self))
    test('5')
    test('5\n\n\nfdsijhfdiosahfdsajfdias')
    test(None)

    test = self.subtest_attribute(TextAttribute, multiline=True)
    test(5, '5')
    test(5.2, '5.2')
    test(self, str(self))
    test('5')
    test('5\n\n\nfdsijhfdiosahfdsajfdias')
    test(None)

    test = self.subtest_attribute(IntegerAttribute)
    test(5)
    test(5.2, 5)
    self.assertRaises(TypeError, test, self)
    test('5', 5)
    self.assertRaises(TypeError, test, '5a')
    test(None)

    test = self.subtest_attribute(FloatAttribute)
    test(5, 5.0)
    test(5.2)
    self.assertRaises(TypeError, test, self)
    test('5', 5.0)
    self.assertRaises(TypeError, test, '5a')
    test(None)

    test = self.subtest_attribute(BooleanAttribute)
    test(5, True)
    test(5.2, True)
    test(self, True)
    test('5', True)
    test('5a', True)
    test(True)
    test(False)
    test(None)


  def test_keyattributes(self):

    test = self.subtest_attribute(KeyAttribute)
    test(5, Key(5))
    test(5.2, Key(5.2))
    test(self, Key(self))
    test('5', Key('5'))
    self.assertRaises(ValueError, test, '5\n\n\nfdsijhfdiosahfdsajfdias')
    test(None)

    test = self.subtest_attribute(KeyAttribute, type='Herp')
    self.assertRaises(ValueError, test, 5)
    self.assertRaises(ValueError, test, 5.2)
    self.assertRaises(ValueError, test, self)
    self.assertRaises(ValueError, test, '5')
    self.assertRaises(ValueError, test, '5\n\n\nfdsijhfdiosahfdsajfdias')
    test(None)
    self.assertRaises(ValueError, test, '/Herp')
    test(Key('/Herp:Derp'))
    self.assertRaises(ValueError, test, '/Herp:Derp/Lerp')

    test = self.subtest_attribute(KeyAttribute, parent='/Herp:Derp')
    self.assertRaises(ValueError, test, 5)
    self.assertRaises(ValueError, test, 5.2)
    self.assertRaises(ValueError, test, self)
    self.assertRaises(ValueError, test, '5')
    self.assertRaises(ValueError, test, '5\n\n\nfdsijhfdiosahfdsajfdias')
    test(None)
    self.assertRaises(ValueError, test, '/Herp')
    self.assertRaises(ValueError, test, '/Herp:Derp')
    test(Key('/Herp:Derp/Lerp'))
    self.assertRaises(ValueError, test, '/Merp/Herp:Derp/Lerp')

    test = self.subtest_attribute(KeyAttribute, ancestor='/Herp')
    self.assertRaises(ValueError, test, 5)
    self.assertRaises(ValueError, test, 5.2)
    self.assertRaises(ValueError, test, self)
    self.assertRaises(ValueError, test, '5')
    self.assertRaises(ValueError, test, '5\n\n\nfdsijhfdiosahfdsajfdias')
    test(None)
    self.assertRaises(ValueError, test, '/Herp')
    test(Key('/Herp/Derp'))
    test(Key('/Herp/Derp/Lerp'))
    self.assertRaises(ValueError, test, '/Merp/Herp/Derp/Lerp')

    test = self.subtest_attribute(KeyAttribute, descendant='/Herp/Derp/Lerp')
    self.assertRaises(ValueError, test, 5)
    self.assertRaises(ValueError, test, 5.2)
    self.assertRaises(ValueError, test, self)
    self.assertRaises(ValueError, test, '5')
    self.assertRaises(ValueError, test, '5\n\n\nfdsijhfdiosahfdsajfdias')
    test(None)
    test(Key('/Herp'))
    test(Key('/Herp/Derp'))
    self.assertRaises(ValueError, test, '/Herp/Derp/Lerp')
    self.assertRaises(ValueError, test, '/Merp/Herp/Derp')
    self.assertRaises(ValueError, test, '/Merp/Herp/Derp/Lerp')


  def test_other(self):

    test = self.subtest_attribute(TimeAttribute)
    test(5, nanotime.nanotime(5))
    test(5.2, nanotime.nanotime(5.2))
    self.assertRaises(TypeError, test, self)
    self.assertRaises(TypeError, test, '5')
    self.assertRaises(TypeError, test, '5a')
    test(nanotime.seconds(1000))
    test(None)

    test = self.subtest_attribute(DateTimeAttribute)
    self.assertRaises(TypeError, test, 5)
    self.assertRaises(TypeError, test, 5.2)
    self.assertRaises(TypeError, test, self)
    self.assertRaises(ValueError, test, '5')
    self.assertRaises(ValueError, test, '5a')
    test(datetime.datetime.now())
    test(None)
    d = datetime.datetime(2011, 7, 12, 13, 19, 29, 151455)
    test(str(d), d)
    d = datetime.datetime(2011, 7, 12, 13, 19, 29, 0)
    test(str(d), d)

    test = self.subtest_attribute(ListAttribute)
    self.assertRaises(TypeError, test, 5, ['5'])
    self.assertRaises(TypeError, test, 5.2, ['5.2'])
    self.assertRaises(TypeError, test, self, [str(self)])
    test('5', ['5'])
    test('5a', ['5', 'a'])
    test([])
    test(['fdgfds', 'gfdsgfds', 'gfdsgfds', 'gfdsgfds'])
    test([4214, 321, 43, 21], ['4214', '321', '43', '21'])
    test(xrange(0, 10), map(str, range(0, 10)))
    test(None)

    test = self.subtest_attribute(DictAttribute)
    self.assertRaises(TypeError, test, 5)
    self.assertRaises(TypeError, test, 5.2)
    self.assertRaises(TypeError, test, self)
    self.assertRaises(TypeError, test, '5')
    self.assertRaises(TypeError, test, '5a')
    test({})
    test({'a':'b'})
    test({'1213':3214}, {'1213':'3214'})
    test({1213:3214}, {'1213':'3214'})
    test(None)

if __name__ == '__main__':
  unittest.main()



import time
import datetime
import unittest
import hashlib
import nanotime

from dronestore.model import Key, Version, Model
from dronestore.query import Filter, Order, Query
from dronestore.util import serial
from dronestore import model


def versions():
  sr = serial.SerialRepresentation()
  sr['key'] = '/ABCD'
  sr['hash'] = hashlib.sha1('herp').hexdigest()
  sr['parent'] = Version.BLANK_HASH
  sr['created'] = nanotime.now().nanoseconds()
  sr['committed'] = nanotime.now().nanoseconds()
  sr['attributes'] = {'str' : {'value' : 'herp'} }
  sr['type'] = 'Hurr'

  v1 = Version(sr)

  sr = serial.SerialRepresentation()
  sr['key'] = '/ABCD'
  sr['hash'] = hashlib.sha1('derp').hexdigest()
  sr['parent'] = hashlib.sha1('herp').hexdigest()
  sr['created'] = nanotime.now().nanoseconds()
  sr['committed'] = nanotime.now().nanoseconds()
  sr['attributes'] = {'str' : {'value' : 'derp'} }
  sr['type'] = 'Hurr'

  v2 = Version(sr)

  sr = serial.SerialRepresentation()
  sr['key'] = '/ABCD'
  sr['hash'] = hashlib.sha1('lerp').hexdigest()
  sr['parent'] = hashlib.sha1('derp').hexdigest()
  sr['created'] = nanotime.now().nanoseconds()
  sr['committed'] = nanotime.now().nanoseconds()
  sr['attributes'] = {'str' : {'value' : 'lerp'} }
  sr['type'] = 'Hurr'

  v3 = Version(sr)

  return v1, v2, v3


class TestFilter(unittest.TestCase):

  def assertEqualFilter(self, af, bf, *args):
    return self.assertEqual([a for a in af], [a for a in bf], *args)

  def test_basic(self):

    v1, v2, v3 = versions()
    vs = [v1, v2, v3]

    t1 = v1.committed
    t2 = v2.committed
    t3 = v3.committed

    fkgtA = Filter('key', '>', '/A')

    self.assertTrue(fkgtA(v1))
    self.assertTrue(fkgtA(v2))
    self.assertTrue(fkgtA(v3))

    self.assertTrue(fkgtA.valuePasses('/BCDEG'))
    self.assertTrue(fkgtA.valuePasses('/ZCDEFDSA/fdsafdsa/fdsafdsaf'))
    self.assertFalse(fkgtA.valuePasses('/6353456346543'))
    self.assertFalse(fkgtA.valuePasses('.'))
    self.assertTrue(fkgtA.valuePasses('afsdafdsa'))

    self.assertEqualFilter(Filter.filter([fkgtA], vs), vs)

    fkltA = Filter('key', '<', '/A')

    self.assertFalse(fkltA(v1))
    self.assertFalse(fkltA(v2))
    self.assertFalse(fkltA(v3))

    self.assertFalse(fkltA.valuePasses('/BCDEG'))
    self.assertFalse(fkltA.valuePasses('/ZCDEFDSA/fdsafdsa/fdsafdsaf'))
    self.assertTrue(fkltA.valuePasses('/6353456346543'))
    self.assertTrue(fkltA.valuePasses('.'))
    self.assertFalse(fkltA.valuePasses('A'))
    self.assertFalse(fkltA.valuePasses('afsdafdsa'))

    self.assertEqualFilter(Filter.filter([fkltA], vs), [])

    fkeqA = Filter('key', '=', '/ABCD')

    self.assertTrue(fkeqA(v1))
    self.assertTrue(fkeqA(v2))
    self.assertTrue(fkeqA(v3))

    self.assertFalse(fkeqA.valuePasses('/BCDEG'))
    self.assertFalse(fkeqA.valuePasses('/ZCDEFDSA/fdsafdsa/fdsafdsaf'))
    self.assertFalse(fkeqA.valuePasses('/6353456346543'))
    self.assertFalse(fkeqA.valuePasses('A'))
    self.assertFalse(fkeqA.valuePasses('.'))
    self.assertFalse(fkeqA.valuePasses('afsdafdsa'))
    self.assertTrue(fkeqA.valuePasses('/ABCD'))

    self.assertEqualFilter(Filter.filter([fkeqA], vs), vs)
    self.assertEqualFilter(Filter.filter([fkeqA, fkltA], vs), [])
    self.assertEqualFilter(Filter.filter([fkgtA, fkeqA], vs), vs)

    fkgtB = Filter('key', '>', '/B')

    self.assertFalse(fkgtB(v1))
    self.assertFalse(fkgtB(v2))
    self.assertFalse(fkgtB(v3))

    self.assertFalse(fkgtB.valuePasses('/A'))
    self.assertTrue(fkgtB.valuePasses('/BCDEG'))
    self.assertTrue(fkgtB.valuePasses('/ZCDEFDSA/fdsafdsa/fdsafdsaf'))
    self.assertFalse(fkgtB.valuePasses('/6353456346543'))
    self.assertFalse(fkgtB.valuePasses('.'))
    self.assertTrue(fkgtB.valuePasses('A'))
    self.assertTrue(fkgtB.valuePasses('afsdafdsa'))

    self.assertEqualFilter(Filter.filter([fkgtB], vs), [])
    self.assertEqualFilter(Filter.filter([fkgtB, fkgtA], vs), [])
    self.assertEqualFilter(Filter.filter([fkgtA, fkgtB], vs), [])

    fkltB = Filter('key', '<', '/B')

    self.assertTrue(fkltB(v1))
    self.assertTrue(fkltB(v2))
    self.assertTrue(fkltB(v3))

    self.assertTrue(fkltB.valuePasses('/A'))
    self.assertFalse(fkltB.valuePasses('/BCDEG'))
    self.assertFalse(fkltB.valuePasses('/ZCDEFDSA/fdsafdsa/fdsafdsaf'))
    self.assertTrue(fkltB.valuePasses('/6353456346543'))
    self.assertTrue(fkltB.valuePasses('.'))
    self.assertFalse(fkltB.valuePasses('A'))
    self.assertFalse(fkltB.valuePasses('afsdafdsa'))

    self.assertEqualFilter(Filter.filter([fkltB], vs), vs)

    fkgtAB = Filter('key', '>', '/AB')

    self.assertTrue(fkgtAB(v1))
    self.assertTrue(fkgtAB(v2))
    self.assertTrue(fkgtAB(v3))

    self.assertFalse(fkgtAB.valuePasses('/A'))
    self.assertTrue(fkgtAB.valuePasses('/BCDEG'))
    self.assertTrue(fkgtAB.valuePasses('/ZCDEFDSA/fdsafdsa/fdsafdsaf'))
    self.assertFalse(fkgtAB.valuePasses('/6353456346543'))
    self.assertFalse(fkgtAB.valuePasses('.'))
    self.assertTrue(fkgtAB.valuePasses('A'))
    self.assertTrue(fkgtAB.valuePasses('afsdafdsa'))

    self.assertEqualFilter(Filter.filter([fkgtAB], vs), vs)
    self.assertEqualFilter(Filter.filter([fkgtAB, fkltB], vs), vs)
    self.assertEqualFilter(Filter.filter([fkltB, fkgtAB], vs), vs)

    fgtet1 = Filter('committed', '>=', t1)
    fgtet2 = Filter('committed', '>=', t2)
    fgtet3 = Filter('committed', '>=', t3)

    self.assertTrue(fgtet1(v1))
    self.assertTrue(fgtet1(v2))
    self.assertTrue(fgtet1(v3))

    self.assertFalse(fgtet2(v1))
    self.assertTrue(fgtet2(v2))
    self.assertTrue(fgtet2(v3))

    self.assertFalse(fgtet3(v1))
    self.assertFalse(fgtet3(v2))
    self.assertTrue(fgtet3(v3))

    self.assertEqualFilter(Filter.filter([fgtet1], vs), vs)
    self.assertEqualFilter(Filter.filter([fgtet2], vs), [v2, v3])
    self.assertEqualFilter(Filter.filter([fgtet3], vs), [v3])

    fltet1 = Filter('committed', '<=', t1)
    fltet2 = Filter('committed', '<=', t2)
    fltet3 = Filter('committed', '<=', t3)

    self.assertTrue(fltet1(v1))
    self.assertFalse(fltet1(v2))
    self.assertFalse(fltet1(v3))

    self.assertTrue(fltet2(v1))
    self.assertTrue(fltet2(v2))
    self.assertFalse(fltet2(v3))

    self.assertTrue(fltet3(v1))
    self.assertTrue(fltet3(v2))
    self.assertTrue(fltet3(v3))

    self.assertEqualFilter(Filter.filter([fltet1], vs), [v1])
    self.assertEqualFilter(Filter.filter([fltet2], vs), [v1, v2])
    self.assertEqualFilter(Filter.filter([fltet3], vs), vs)

    self.assertEqualFilter(Filter.filter([fgtet2, fltet2], vs), [v2])
    self.assertEqualFilter(Filter.filter([fgtet1, fltet3], vs), vs)
    self.assertEqualFilter(Filter.filter([fgtet3, fltet1], vs), [])

    feqt1 = Filter('committed', '=', t1)
    feqt2 = Filter('committed', '=', t2)
    feqt3 = Filter('committed', '=', t3)

    self.assertTrue(feqt1(v1))
    self.assertFalse(feqt1(v2))
    self.assertFalse(feqt1(v3))

    self.assertFalse(feqt2(v1))
    self.assertTrue(feqt2(v2))
    self.assertFalse(feqt2(v3))

    self.assertFalse(feqt3(v1))
    self.assertFalse(feqt3(v2))
    self.assertTrue(feqt3(v3))

    self.assertEqualFilter(Filter.filter([feqt1], vs), [v1])
    self.assertEqualFilter(Filter.filter([feqt2], vs), [v2])
    self.assertEqualFilter(Filter.filter([feqt3], vs), [v3])


  def test_object(self):
    t1 = nanotime.now()
    t2 = nanotime.now()

    f1 = Filter('key', '>', '/A')
    f2 = Filter('key', '<', '/A')
    f3 = Filter('committed', '=', t1)
    f4 = Filter('committed', '>=', t2)

    self.assertEqual(f1, eval(repr(f1)))
    self.assertEqual(f2, eval(repr(f2)))
    self.assertEqual(f3, eval(repr(f3)))
    self.assertEqual(f4, eval(repr(f4)))

    self.assertEqual(str(f1), 'key > /A')
    self.assertEqual(str(f2), 'key < /A')
    self.assertEqual(str(f3), 'committed = %s' % t1)
    self.assertEqual(str(f4), 'committed >= %s' % t2)

    self.assertEqual(f1, Filter('key', '>', '/A'))
    self.assertEqual(f2, Filter('key', '<', '/A'))
    self.assertEqual(f3, Filter('committed', '=', t1))
    self.assertEqual(f4, Filter('committed', '>=', t2))

    self.assertNotEqual(f2, Filter('key', '>', '/A'))
    self.assertNotEqual(f1, Filter('key', '<', '/A'))
    self.assertNotEqual(f4, Filter('committed', '=', t1))
    self.assertNotEqual(f3, Filter('committed', '>=', t2))

    self.assertEqual(hash(f1), hash(Filter('key', '>', '/A')))
    self.assertEqual(hash(f2), hash(Filter('key', '<', '/A')))
    self.assertEqual(hash(f3), hash(Filter('committed', '=', t1)))
    self.assertEqual(hash(f4), hash(Filter('committed', '>=', t2)))

    self.assertNotEqual(hash(f2), hash(Filter('key', '>', '/A')))
    self.assertNotEqual(hash(f1), hash(Filter('key', '<', '/A')))
    self.assertNotEqual(hash(f4), hash(Filter('committed', '=', t1)))
    self.assertNotEqual(hash(f3), hash(Filter('committed', '>=', t2)))



class TestOrder(unittest.TestCase):

  def test_basic(self):
    o1 = Order('key')
    o2 = Order('+committed')
    o3 = Order('-created')

    v1, v2, v3 = versions()

    # test  isAscending
    self.assertTrue(o1.isAscending())
    self.assertTrue(o2.isAscending())
    self.assertFalse(o3.isAscending())

    # test keyfn
    self.assertEqual(o1.keyfn(v1), (v1.key))
    self.assertEqual(o1.keyfn(v2), (v2.key))
    self.assertEqual(o1.keyfn(v3), (v3.key))
    self.assertEqual(o1.keyfn(v1), (v2.key))
    self.assertEqual(o1.keyfn(v1), (v3.key))

    self.assertEqual(o2.keyfn(v1), (v1.committed))
    self.assertEqual(o2.keyfn(v2), (v2.committed))
    self.assertEqual(o2.keyfn(v3), (v3.committed))
    self.assertNotEqual(o2.keyfn(v1), (v2.committed))
    self.assertNotEqual(o2.keyfn(v1), (v3.committed))

    self.assertEqual(o3.keyfn(v1), (v1.created))
    self.assertEqual(o3.keyfn(v2), (v2.created))
    self.assertEqual(o3.keyfn(v3), (v3.created))
    self.assertNotEqual(o3.keyfn(v1), (v2.created))
    self.assertNotEqual(o3.keyfn(v1), (v3.created))

    # test sorted
    self.assertEqual(Order.sorted([v3, v2, v1], [o1]), [v3, v2, v1])
    self.assertEqual(Order.sorted([v3, v2, v1], [o1, o2]), [v1, v2, v3])
    self.assertEqual(Order.sorted([v1, v3, v2], [o1, o3]), [v3, v2, v1])
    self.assertEqual(Order.sorted([v3, v2, v1], [o1, o2, o3]), [v1, v2, v3])
    self.assertEqual(Order.sorted([v1, v3, v2], [o1, o3, o2]), [v3, v2, v1])

    self.assertEqual(Order.sorted([v3, v2, v1], [o2]), [v1, v2, v3])
    self.assertEqual(Order.sorted([v3, v2, v1], [o2, o1]), [v1, v2, v3])
    self.assertEqual(Order.sorted([v3, v2, v1], [o2, o3]), [v1, v2, v3])
    self.assertEqual(Order.sorted([v3, v2, v1], [o2, o1, o3]), [v1, v2, v3])
    self.assertEqual(Order.sorted([v3, v2, v1], [o2, o3, o1]), [v1, v2, v3])

    self.assertEqual(Order.sorted([v1, v2, v3], [o3]), [v3, v2, v1])
    self.assertEqual(Order.sorted([v1, v2, v3], [o3, o2]), [v3, v2, v1])
    self.assertEqual(Order.sorted([v1, v2, v3], [o3, o1]), [v3, v2, v1])
    self.assertEqual(Order.sorted([v1, v2, v3], [o3, o2, o1]), [v3, v2, v1])
    self.assertEqual(Order.sorted([v1, v2, v3], [o3, o1, o2]), [v3, v2, v1])


  def test_object(self):
    self.assertEqual(Order('key'), eval(repr(Order('key'))))
    self.assertEqual(Order('+committed'), eval(repr(Order('+committed'))))
    self.assertEqual(Order('-created'), eval(repr(Order('-created'))))

    self.assertEqual(str(Order('key')), '+key')
    self.assertEqual(str(Order('+committed')), '+committed')
    self.assertEqual(str(Order('-created')), '-created')

    self.assertEqual(Order('key'), Order('+key'))
    self.assertEqual(Order('-key'), Order('-key'))
    self.assertEqual(Order('+committed'), Order('+committed'))

    self.assertNotEqual(Order('key'), Order('-key'))
    self.assertNotEqual(Order('+key'), Order('-key'))
    self.assertNotEqual(Order('+committed'), Order('+key'))

    self.assertEqual(hash(Order('+key')), hash(Order('+key')))
    self.assertEqual(hash(Order('-key')), hash(Order('-key')))
    self.assertNotEqual(hash(Order('+key')), hash(Order('-key')))
    self.assertEqual(hash(Order('+committed')), hash(Order('+committed')))
    self.assertNotEqual(hash(Order('+committed')), hash(Order('+key')))


class TestQuery(unittest.TestCase):

  def test_basic(self):
    fn = lambda: Query('UnregistedModel').model()
    self.assertRaises(model.UnregisteredModelError, fn)

    now = nanotime.now().nanoseconds()

    q1 = Query('Model', limit=100)
    q2 = Query('Model', offset=200)
    q3 = Query('Model')

    q1.offset = 300
    q3.limit = 1

    q1.filter('key', '>', '/ABC')
    q1.filter('created', '>', now)

    q2.order('key')
    q2.order('-created')

    q1d = {'key' : '/Model', 'limit':100, 'offset':300, \
      'filter': [['key', '>', '/ABC'], ['created', '>', now]] }

    q2d = {'key' : '/Model', 'offset':200, 'order': ['+key', '-created'] }

    q3d = {'key' : '/Model', 'limit':1}

    self.assertEqual(q1.dict(), q1d)
    self.assertEqual(q2.dict(), q2d)
    self.assertEqual(q3.dict(), q3d)

    self.assertEqual(q1, Query.from_dict(q1d))
    self.assertEqual(q2, Query.from_dict(q2d))
    self.assertEqual(q3, Query.from_dict(q3d))

    self.assertEqual(q1, eval(repr(q1)))
    self.assertEqual(q2, eval(repr(q2)))
    self.assertEqual(q3, eval(repr(q3)))


if __name__ == '__main__':
  unittest.main()




import datetime
import nanotime

import merge
import model

class Attribute(object):
  '''Attributes define and compose a Model. A Model can be seen as a collection
  of attributes.

  An Attribute primarily defines a name, an associated data type, and a
  particular merge strategy.

  Attributes can have other options, including defining a default value, and
  validation for the data they hold.
  '''
  data_type = str
  default_strategy = merge.LatestObjectStrategy

  def __init__(self, name=None, default=None, required=False, strategy=None):

    if not strategy:
      strategy = self.default_strategy
    strategy = strategy(self)

    if not isinstance(strategy, merge.MergeStrategy):
      raise TypeError('mergeStrategy does not inherit from %s' % \
        merge.MergeStrategy)

    strategy.attribute = self

    self.name = name
    self.default = default
    self.required = required
    self.mergeStrategy = strategy


  def _attr_config(self, model_class, attr_name):
    '''Configure attribute for a given model.'''
    self.__model__ = model_class
    if self.name is None:
      self.name = attr_name

  def _attr_name(self):
    '''Returns the attribute name within the model instance.'''
    return '_' + self.name

  def rawData(self, instance):
    if instance is None:
      return None

    try:
      return getattr(instance, self._attr_name())
    except AttributeError:
      return None

  def setRawData(self, instance, rawData):
    setattr(instance, self._attr_name(), rawData)
    instance._isDirty = True

  def __get__(self, instance, model_class):
    '''Descriptor to aid model instantiation.'''
    if instance is None:
      return self

    try:
      rawData = getattr(instance, self._attr_name())
      return self.loads(rawData['value'])
    except AttributeError:
      return self.default_value()

  def __set__(self, instance, value, default=False):
    '''Validate and Set the attribute on the model instance.'''
    if not default:
      value = self.validate(value)

    rawData = self.rawData(instance)
    if rawData is None:
      rawData = {}
      setattr(instance, self._attr_name(), rawData)

    # our attributes are idempotent, so if its the same, doesn't change state
    if 'value' in rawData:
      oldval = rawData['value']
      if value is None and oldval is None:
        return
      if value is not None and oldval is not None and oldval == value:
        return

    rawData['value'] = self.dumps(value)
    instance._isDirty = True
    self.mergeStrategy.setAttribute(instance, rawData, default=default)

  def default_value(self):
    '''The default value for a particular attribute.'''
    return self.default

  def validate(self, value):
    '''Assert that the provided value is compatible with this attribute.'''
    if self.empty(value):
      if self.required:
        raise ValueError('Attribute %s is required.' % self.name)

    if value is not None and not isinstance(value, self.data_type):
      try:
        value = self.data_type(value)
      except:
        errstr = 'value for attribute %s is of type %s, not %s'
        raise TypeError(errstr % (self.name, type(value), self.data_type))

    return value

  def empty(self, value):
    '''Simple check to determine if value is empty.'''
    return not value

  # the loads/dumps interface here is to provide an easy way for attributes
  # to store serialized information.
  def dumps(self, value):
    '''Returns value into raw data to store.'''
    return value

  def loads(self, raw):
    '''Converts raw data into value.'''
    return raw


class StringAttribute(Attribute):
  '''Keep compatibility with App Engine by using basestrings as well'''
  data_type = basestring

  def __init__(self, multiline=False, **kwds):
    super(StringAttribute, self).__init__(**kwds)
    self.multiline = multiline

  def validate(self, value):
    if value is not None and not isinstance(value, self.data_type):
      value = str(value)

    value = super(StringAttribute, self).validate(value)

    if not self.multiline and value and '\n' in value:
      raise ValueError('Attribute %s is not multi-line' % self.name)

    return value


class KeyAttribute(StringAttribute):
  '''Attribute to store Keys.'''
  data_type = model.Key

  def __init__(self, type=None, parent=None, ancestor=None, \
    descendant=None, **kwds):
    ktype = type
    type = self.__class__.__class__

    if 'multiline' not in kwds:
      kwds['multiline'] = False
    super(KeyAttribute, self).__init__(**kwds)

    if ktype and isinstance(ktype, type) and issubclass(ktype, model.Model):
      ktype = ktype.__dstype__

    self.type = str(ktype) if ktype else None
    self.parent = model.Key(parent) if parent else None
    self.ancestor = model.Key(ancestor) if ancestor else None
    self.descendant = model.Key(descendant) if descendant else None

  def validate(self, value):
    '''Ensures key value matches criteria.'''
    value = super(KeyAttribute, self).validate(value)
    errstr = "key '%s' for attribute '%s'" % (value, self.name)

    if value is not None:

      # make sure the key type matches.
      if self.type is not None and value.type != self.type:
        errstr += " must be of key type '%s'" % self.type
        raise ValueError(errstr)

      # make sure the parent key matches.
      if self.parent and value.parent != self.parent:
        errstr += " must have parent key '%s'" % self.parent
        raise ValueError(errstr)

      # make sure the ancestry matches.
      if self.ancestor and not self.ancestor.isAncestorOf(value):
        errstr += "must have ancestor key '%s'" % self.ancestor
        raise ValueError(errstr)

      # make sure the ancestry matches.
      if self.descendant and not self.descendant.isDescendantOf(value):
        errstr = " must be a descendant of key '%s'" % self.descendant
        raise ValueError(errstr)

    return value



class TextAttribute(StringAttribute):
  '''Attribute to store large amounts of text. Datastores should optimize.'''

  def __init__(self, **kwds):
    if 'multiline' not in kwds:
      kwds['multiline'] = True
    super(TextAttribute, self).__init__(**kwds)



class IntegerAttribute(Attribute):
  '''Integer Attribute'''
  data_type = int

  def validate(self, value):
    value = super(IntegerAttribute, self).validate(value)
    if value is None:
      return value

    if not isinstance(value, (int, long)) or isinstance(value, bool):
      raise ValueError('Attribute %s must be an int or long, not a %s'
                          % (self.name, type(value).__name__))

    if value < -0x8000000000000000 or value > 0x7fffffffffffffff:
      raise ValueError('Property %s must fit in 64 bits' % self.name)

    return value

  def empty(self, value):
    '''0 is not empty.'''
    return value is None



class FloatAttribute(Attribute):
  '''Floating point Attribute'''
  data_type = float

  def empty(self, value):
    '''0 is not empty.'''
    return value is None


class BooleanAttribute(Attribute):
  '''Boolean Attribute'''
  data_type = bool

  def empty(self, value):
    '''False is not empty.'''
    return value is None



class TimeAttribute(Attribute):
  '''Attribute to store nanosecond times.'''
  data_type = nanotime.nanotime

  # store the data as nanoseconds
  @classmethod
  def dumps(cls, nanotime_):
    if nanotime_ is None:
      return None
    return nanotime_.nanoseconds()

  @classmethod
  def loads(cls, nanoseconds):
    if nanoseconds is None:
      return None
    return nanotime.nanoseconds(nanoseconds)




class DateTimeAttribute(TimeAttribute):
  '''Attribute to store nanosecond times and return datetime objects.'''

  data_type = datetime.datetime

  @classmethod
  def dumps(cls, datetime_):
    if datetime_ is None:
      return None
    return datetime_.isoformat()

  @classmethod
  def loads(cls, iso_string):
    if iso_string is None:
      return None
    return cls._datetime_from_iso_string(iso_string)

  def __set__(self, instance, value, default=False):
    '''Set the attribute on the model instance.'''

    if isinstance(value, basestring):
      value = self._datetime_from_iso_string(value)

    super(DateTimeAttribute, self).__set__(instance, value, default=default)

  def empty(self, value):
    '''0 is not empty.'''
    return value is None


  @classmethod
  def _datetime_from_iso_string(cls, value):
    microseconds = 0
    if value and '.' in value:
      value, frag = value.rsplit('.', 1)
      frag = frag[:6]  # truncate to microseconds
      frag = frag.replace('Z', '')
      frag += (6 - len(frag)) * '0'  # add 0s
      microseconds = int(frag)

    sep = 'T' if 'T' in value else ' '
    fmt = '%Y-%m-%d' + sep + '%H:%M:%S'

    value = datetime.datetime.strptime(value, fmt)
    value = value.replace(microsecond=microseconds)
    return value




class ListAttribute(Attribute):
  '''Attribute to store lists.'''
  data_type = list
  data_value_type = str

  def __init__(self, value_type=None, **kwds):
    super(ListAttribute, self).__init__(**kwds)
    if value_type:
      self.data_value_type = value_type

  def validate(self, value):
    value = super(ListAttribute, self).validate(value)
    if value is None:
      return value

    for i in xrange(0, len(value)):
      val = value[i]
      if not isinstance(val, self.data_value_type):
        try:
          value[i] = self.data_value_type(val)
        except:
          errstr = 'internal value for attribute %s is not of type %s'
          raise TypeError(errstr % (self.name, self.data_value_type))

    return value

  def empty(self, value):
    '''[] is not empty.'''
    return value is None





class DictAttribute(ListAttribute):
  '''Attribute to store lists.'''
  data_type = dict
  data_value_type = str

  def validate(self, value):
    value = super(ListAttribute, self).validate(value)
    if value is None:
      return value

    for key, val in value.items():

      # Make sure all keys are strings
      if not isinstance(key, basestring):
        try:
          value[str(key)] = value[key]
          del value[key]
          key = str(key)
        except:
          errstr = 'internal key for attribute %s must be a string'
          raise TypeError(errstr % self.name)

      # Make sure all values are of type `data_value_type`
      if not isinstance(val, self.data_value_type):
        try:
          value[key] = self.data_value_type(val)
        except:
          errstr = 'internal value for attribute %s must be of type %s'
          raise TypeError(errstr % (self.name, self.data_value_type))

    return value

  def empty(self, value):
    '''{} is not empty.'''
    return value is None






import unittest
import hashlib
import nanotime

from dronestore.util import serial
from dronestore.util import fasthash
from dronestore.model import *
from dronestore.attribute import *

from .util import RandomGen


class Person(Model):
  first = StringAttribute(default="Firstname")
  last = StringAttribute(default="Lastname")
  phone = StringAttribute(default="N/A")
  age = IntegerAttribute(default=0)
  gender = StringAttribute()

  def validate(self):
    super(Person, self).validate()

    if self.gender == self.first:
      raise ValueError('Sorry, a gender is not an acceptable first name!')



class KeyTests(unittest.TestCase):

  def __subtest_basic(self, string):
    fixedString = Key.removeDuplicateSlashes(string)
    self.assertEqual(Key(string)._string, fixedString)
    self.assertEqual(Key(string), Key(string))
    self.assertEqual(str(Key(string)), fixedString)
    self.assertEqual(repr(Key(string)), "Key('%s')" % fixedString)
    self.assertEqual(Key(string).name, fixedString.rsplit('/')[-1])
    self.assertEqual(Key(string), eval(repr(Key(string))))

    self.assertRaises(TypeError, cmp, Key(string), string)

    split = fixedString.split('/')
    self.assertEqual(Key('/'.join(split[:-1])), Key(string).parent)

  def test_basic(self):
    self.__subtest_basic('')
    self.__subtest_basic('abcde')
    self.__subtest_basic('disahfidsalfhduisaufidsail')
    self.__subtest_basic('/fdisahfodisa/fdsa/fdsafdsafdsafdsa/fdsafdsa/')
    self.__subtest_basic(u'4215432143214321432143214321')
    self.__subtest_basic('/fdisaha////fdsa////fdsafdsafdsafdsa/fdsafdsa/')


  def test_ancestry(self):
    k1 = Key('/A/B/C')
    k2 = Key('/A/B/C/D')

    self.assertEqual(k1._string, '/A/B/C')
    self.assertEqual(k2._string, '/A/B/C/D')
    self.assertTrue(k1.isAncestorOf(k2))
    self.assertTrue(k2.isDescendantOf(k1))
    self.assertTrue(Key('/A').isAncestorOf(k2))
    self.assertTrue(Key('/A').isAncestorOf(k1))
    self.assertFalse(Key('/A').isDescendantOf(k2))
    self.assertFalse(Key('/A').isDescendantOf(k1))
    self.assertTrue(k2.isDescendantOf(Key('/A')))
    self.assertTrue(k1.isDescendantOf(Key('/A')))
    self.assertFalse(k2.isAncestorOf(Key('/A')))
    self.assertFalse(k1.isAncestorOf(Key('/A')))
    self.assertFalse(k2.isAncestorOf(k2))
    self.assertFalse(k1.isAncestorOf(k1))
    self.assertEqual(k1.child('D'), k2)
    self.assertEqual(k1, k2.parent)

    self.assertRaises(TypeError, k1.isAncestorOf, str(k2))
    self.assertEqual(k1.parent, Key('/A/B'))
    self.assertEqual(k2.parent, Key('/A/B/C'))
    self.assertEqual(k2.parent, k1)

  def test_hashing(self):

    def randomKey():
      return Key('/herp/' + RandomGen.randomString() + '/derp')

    keys = {}

    for i in range(0, 200):
      key = randomKey()
      while key in keys.values():
        key = randomKey()

      hstr = str(hash(key))
      self.assertFalse(hstr in keys)
      keys[hstr] = key

    for key in keys.values():
      hstr = str(hash(key))
      self.assertTrue(hstr in keys)
      self.assertEqual(key, keys[hstr])

  def test_random(self):
    keys = set()
    for i in range(0, 1000):
      random = Key.randomKey()
      self.assertFalse(random in keys)
      keys.add(random)
    self.assertEqual(len(keys), 1000)





class VersionTests(unittest.TestCase):

  def test_blank(self):
    blank = Version(Key('/BLANK'))
    self.assertEqual(blank.hash, Version.BLANK_HASH)
    self.assertEqual(blank.type, '')
    self.assertEqual(blank.shortHash(5), Version.BLANK_HASH[0:5])
    self.assertEqual(blank.committed, nanotime.nanotime(0))
    self.assertEqual(blank.created, nanotime.nanotime(0))
    self.assertEqual(blank.parent, Version.BLANK_HASH)

    self.assertEqual(blank, Version(Key('/BLANK')))
    self.assertTrue(blank.isBlank)


  def test_creation(self):

    h1 = hashlib.sha1('derp').hexdigest()
    h2 = hashlib.sha1('herp').hexdigest()
    now = nanotime.now()

    sr = serial.SerialRepresentation()
    sr['key'] = '/A'
    sr['hash'] = h1
    sr['parent'] = h2
    sr['created'] = now.nanoseconds()
    sr['committed'] = now.nanoseconds()
    sr['attributes'] = {'str' : {'value' : 'derp'} }
    sr['type'] = 'Hurr'

    v = Version(sr)
    self.assertEqual(v.type, 'Hurr')
    self.assertEqual(v.hash, h1)
    self.assertEqual(v.parent, h2)
    self.assertEqual(v.created, now)
    self.assertEqual(v.committed, now)
    self.assertEqual(v.shortHash(5), h1[0:5])
    self.assertEqual(v.attributeValue('str'), 'derp')
    self.assertEqual(v.attribute('str')['value'], 'derp')
    self.assertEqual(v['str']['value'], 'derp')
    self.assertEqual(hash(v), hash(fasthash.hash(h1)))
    self.assertEqual(v, Version(sr))
    self.assertFalse(v.isBlank)

    self.assertRaises(KeyError, v.attribute, 'fdsafda')
    self.assertRaises(TypeError, cmp, v, 'fdsafda')

  def test_raises(self):

    sr = serial.SerialRepresentation()
    self.assertRaises(ValueError, Version, sr)
    sr['key'] = '/A'
    self.assertRaises(ValueError, Version, sr)
    sr['hash'] = 'a'
    self.assertRaises(ValueError, Version, sr)
    sr['parent'] = 'b'
    self.assertRaises(ValueError, Version, sr)
    sr['created'] = nanotime.now().nanoseconds()
    self.assertRaises(ValueError, Version, sr)
    sr['committed'] = 0
    self.assertRaises(ValueError, Version, sr)
    sr['committed'] = sr['created']
    self.assertRaises(ValueError, Version, sr)
    sr['attributes'] = {'str' : 'derp'}
    self.assertRaises(ValueError, Version, sr)
    sr['type'] = 'Hurr'
    Version(sr)


  def test_model(self):

    h1 = hashlib.sha1('derp').hexdigest()
    h2 = hashlib.sha1('herp').hexdigest()

    attrs = {'first' : {'value':'Herp'}, \
             'last' : {'value':'Derp'}, \
             'phone' : {'value': '123'}, \
             'age' : {'value': 19}, \
             'gender' : {'value' : 'Male'}}

    sr = serial.SerialRepresentation()
    sr['key'] = '/Person:PersonA'
    sr['hash'] = h1
    sr['parent'] = h2
    sr['created'] = nanotime.now().nanoseconds()
    sr['committed'] = sr['created']
    sr['attributes'] = attrs
    sr['type'] = 'Person'

    ver = Version(sr)
    instance = Person(ver)

    self.assertEqual(instance.__dstype__, ver.type)
    self.assertEqual(instance.version, ver)
    self.assertFalse(instance.isDirty())
    self.assertTrue(instance.isPersisted())
    self.assertTrue(instance.isCommitted())

    self.assertEqual(instance.key, Key('/Person:PersonA'))
    self.assertEqual(instance.first, 'Herp')
    self.assertEqual(instance.last, 'Derp')
    self.assertEqual(instance.phone, '123')
    self.assertEqual(instance.age, 19)
    self.assertEqual(instance.gender, 'Male')


class ModelTests(unittest.TestCase):

  def subtest_assert_uncommitted(self, instance):
    self.assertTrue(instance.created == 0)
    self.assertTrue(instance.committed == 0)
    self.assertTrue(instance.version.isBlank)

    self.assertTrue(instance.isDirty())
    self.assertFalse(instance.isPersisted())
    self.assertFalse(instance.isCommitted())

  def test_basic(self):

    now = nanotime.now()

    a = Model('A')
    self.assertEqual(Model.key, Key('/Model'))
    self.assertEqual(a.key, Key('/Model:A'))
    self.assertEqual(Model.key, a.key.path)
    self.assertEqual(Model.key.instance('A'), a.key)
    self.assertEqual(a.__dstype__, 'Model')
    self.assertEqual(Model.__dstype__, 'Model')
    self.subtest_assert_uncommitted(a)

    a.commit()
    created = a.version.created

    print 'committed', a.version.hash

    self.assertFalse(a.isDirty())
    self.assertTrue(a.isCommitted())
    self.assertEqual(a.version.type, Model.__dstype__)
    self.assertEqual(a.version.hash, a.computedHash())
    self.assertEqual(a.version.parent, Version.BLANK_HASH)
    self.assertEqual(a.version.created, created)
    self.assertEqual(a.created, a.version.created)
    self.assertEqual(a.committed, a.version.committed)
    self.assertTrue(a.created > now)
    self.assertTrue(a.committed > now)

    now = nanotime.now()
    a.commit()

    self.assertFalse(a.isDirty())
    self.assertTrue(a.isCommitted())
    self.assertEqual(a.version.type, Model.__dstype__)
    self.assertEqual(a.version.hash, a.computedHash())
    self.assertEqual(a.version.parent, Version.BLANK_HASH)
    self.assertEqual(a.version.created, created)
    self.assertEqual(a.created, a.version.created)
    self.assertEqual(a.committed, a.version.committed)
    self.assertTrue(a.created < now)
    self.assertTrue(a.committed < now) # didn't REALLY commit.


    a._isDirty = True
    self.assertTrue(a.isDirty())

    now = nanotime.now()
    a.commit()

    self.assertFalse(a.isDirty())
    self.assertTrue(a.isCommitted())
    self.assertEqual(a.version.type, Model.__dstype__)
    self.assertEqual(a.version.hash, a.computedHash())
    self.assertEqual(a.version.parent, Version.BLANK_HASH)
    self.assertEqual(a.version.created, created)
    self.assertEqual(a.created, a.version.created)
    self.assertEqual(a.committed, a.version.committed)
    self.assertTrue(a.created < now)
    self.assertTrue(a.committed < now) # didn't REALLY commit.


  def test_attributes(self):
    p = Person('HerpDerp')
    self.assertEqual(p.key, Key('/Person:HerpDerp'))
    self.assertEqual(p.first, 'Firstname')
    self.assertEqual(p.last, 'Lastname')
    self.assertEqual(p.phone, 'N/A')
    self.assertEqual(p.age, 0)
    self.assertEqual(p.gender, None)

    self.subtest_assert_uncommitted(p)

    p.first = 'Herp'
    p.last = 'Derp'
    p.phone = '1235674444'
    p.age = 120

    p.commit()
    print 'committed', p.version.shortHash(8)

    self.assertFalse(p.isDirty())
    self.assertTrue(p.isCommitted())
    self.assertEqual(p.version.type, Person.__dstype__)
    self.assertEqual(p.version.hash, p.computedHash())
    self.assertEqual(p.version.parent, Version.BLANK_HASH)

    self.assertEqual(p.first, 'Herp')
    self.assertEqual(p.last, 'Derp')
    self.assertEqual(p.phone, '1235674444')
    self.assertEqual(p.age, 120)
    self.assertEqual(p.gender, None)

    self.assertEqual(p.version.attributeValue('first'), 'Herp')
    self.assertEqual(p.version.attributeValue('last'), 'Derp')
    self.assertEqual(p.version.attributeValue('phone'), '1235674444')
    self.assertEqual(p.version.attributeValue('age'), 120)
    self.assertEqual(p.version.attributeValue('gender'), None)

    hash = p.version.hash
    p.first = 'Herpington'
    p.gender = 'Troll'
    p.commit()
    print 'committed', p.version.shortHash(8)

    self.assertFalse(p.isDirty())
    self.assertTrue(p.isCommitted())
    self.assertEqual(p.version.type, Person.__dstype__)
    self.assertEqual(p.version.hash, p.computedHash())
    self.assertEqual(p.version.parent, hash)

    self.assertEqual(p.first, 'Herpington')
    self.assertEqual(p.last, 'Derp')
    self.assertEqual(p.phone, '1235674444')
    self.assertEqual(p.age, 120)
    self.assertEqual(p.gender, 'Troll')

    self.assertEqual(p.version.attributeValue('first'), 'Herpington')
    self.assertEqual(p.version.attributeValue('last'), 'Derp')
    self.assertEqual(p.version.attributeValue('phone'), '1235674444')
    self.assertEqual(p.version.attributeValue('age'), 120)
    self.assertEqual(p.version.attributeValue('gender'), 'Troll')

    p.first = p.gender
    self.assertRaises(ValueError, p.commit)

    class Herp(Model):
      required = StringAttribute(required=True)

    h = Herp('Derp')
    self.assertRaises(ValueError, p.validate)
    self.assertRaises(ValueError, p.commit)

if __name__ == '__main__':
  unittest.main()




__author__ = 'Juan Batiz-Benet'
__email__ = 'juan@benet.ai'
__version__ = '0.2.6'
# don't forget to update setup.py


# basic data model
from model import Key
from model import Version
from model import Model

# attributes
from attribute import Attribute
from attribute import StringAttribute
from attribute import KeyAttribute
from attribute import TextAttribute
from attribute import IntegerAttribute
from attribute import FloatAttribute
from attribute import BooleanAttribute
from attribute import TimeAttribute
from attribute import DateTimeAttribute
from attribute import ListAttribute
from attribute import DictAttribute

# merge strategies
from merge import MergeDirection
from merge import MergeStrategy
from merge import LatestObjectStrategy
from merge import LatestStrategy
from merge import MaxStrategy

# repo
from repo import Repo

# deprecated aliases
drone = repo
Drone = Repo

# query
from query import Query

# basic datastores
from datastore import Datastore
from datastore import DictDatastore

# util
from util.serial import SerialRepresentation


#!/usr/bin/env python

from setuptools import setup, find_packages

import re
main_py = open('dronestore/__init__.py').read()
metadata = dict(re.findall("__([a-z]+)__ = '([^']+)'", main_py))
packages = filter(lambda p: p.startswith('dronestore'), find_packages())

setup(
  name="dronestore",
  version=metadata['version'],
  description="DroneStore python implementation <http://dronestore.org/>",
  author=metadata['author'],
  author_email=metadata['email'],
  url="http://github.com/jbenet/py-dronestore",
  keywords=["dronestore", "data versioning"],
  packages=packages,
  install_requires=[
    "bson>=0.3.3",
    "datastore>=0.2.8",
    "nanotime>=0.5.2",
    "smhasher>=0.136.2",
  ],
  license="MIT License"
)




import unittest
import dronestore
import bottle

from bottledronestore import Plugin

class DronestorePluginTest(unittest.TestCase):
  def setUp(self):
    self.app = bottle.Bottle(catchall=False)

  def test_with_keyword(self):
    self.repo = dronestore.Repo('/Repo')
    self.plugin = self.app.install(Plugin(self.repo))

    @self.app.get('/')
    def test(repo):
      self.assertTrue(isinstance(repo, dronestore.Repo))
      self.assertEqual(self.repo, repo)
    self.app({'PATH_INFO':'/', 'REQUEST_METHOD':'GET'}, lambda x, y: None)

  def test_with_parameter(self):
    self.repo = dronestore.Repo('/Repo')
    self.plugin = self.app.install(Plugin(self.repo))

    @self.app.get('/:param')
    def test(repo, param):
      self.assertTrue(isinstance(repo, dronestore.Repo))
      self.assertEqual(self.repo, repo)
    self.app({'PATH_INFO':'/3', 'REQUEST_METHOD':'GET'}, lambda x, y: None)

  def test_without_keyword(self):
    self.repo = dronestore.Repo('/Repo')
    self.plugin = self.app.install(Plugin(self.repo))

    @self.app.get('/')
    def test():
        pass
    self.app({'PATH_INFO':'/', 'REQUEST_METHOD':'GET'}, lambda x, y: None)

    @self.app.get('/2')
    def test(**kw):
        self.assertFalse('repo' in kw)
    self.app({'PATH_INFO':'/2', 'REQUEST_METHOD':'GET'}, lambda x, y: None)


if __name__ == '__main__':
  unittest.main()



import nanotime
import datetime

def clean(value):
  '''Cleans up a value for insertion into a SerialRepresentation'''

  if isinstance(value, dict):
    value = dict([(clean(k), clean(v)) for k, v in value.items()])
  elif isinstance(value, list) or isinstance(value, tuple):
    value = [clean(v) for v in value]
  elif isinstance(value, int) or isinstance(value, long):
    pass
  elif isinstance(value, float):
    pass
  elif isinstance(value, str):
    pass
  elif isinstance(value, unicode):
    pass
  elif isinstance(value, bool):
    pass
  elif isinstance(value, nanotime.nanotime):
    value = value.nanoseconds()
  elif isinstance(value, datetime.datetime):
    pass
  elif value is None:
    pass
  else: # catch all... turn it into a string!
    value = str(value)
  return value


class SerialRepresentation(object):

  def __init__(self, data=None):
    # internal representation is a dict.
    self._data = clean(data) if data else {}
    self._dirty = True


  def __getitem__(self, key):
    return self._data[key]

  def __setitem__(self, key, data):
    key = clean(key)
    data = clean(data)
    if not self._dirty and self._data[key] == data:
      return # idempotent

    self._dirty = True
    self._data[key] = data

  def __delitem__(self, key):
    self._dirty = True
    del self._data[key]


  def __iter__(self):
    return iter(self._data)

  def __len__(self):
    return len(self._data)

  def __reversed__(self):
    return reversed(self._data)

  def __contains__(self, key):
    return key in self._data

  def __cmp__(self, other):
    return cmp(self._data, other._data)


  def __repr__(self):
    return 'SerialRepresentation(%s)' % repr(self.data())

  def __str__(self):
    return str(self.json())


  def data(self):
    return self._data
